import contextlib
import inspect
import typing

import starlette.types
from di.api.dependencies import DependantBase
from di.container import Container, ContainerState, bind_by_type
from di.dependant import Dependant, JoinedDependant
from di.executors import AsyncExecutor
from starlette.background import BackgroundTasks
from starlette.middleware import Middleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.requests import HTTPConnection, Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.routing import BaseRoute
from starlette.routing import Route as StarletteRoute
from starlette.websockets import WebSocket

from xpresso._utils.asgi import XpressoHTTPExtension, XpressoWebSocketExtension
from xpresso._utils.overrides import DependencyOverrideManager
from xpresso._utils.routing import visit_routes
from xpresso.dependencies.models import Depends, Scopes
from xpresso.exception_handlers import (
    http_exception_handler,
    validation_exception_handler,
)
from xpresso.exceptions import HTTPException, RequestValidationError
from xpresso.middleware.exceptions import ExceptionMiddleware
from xpresso.openapi import models as openapi_models
from xpresso.openapi._builder import genrate_openapi
from xpresso.openapi._html import get_swagger_ui_html
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router
from xpresso.routing.websockets import WebSocketRoute

ExceptionHandler = typing.Callable[
    [Request, Exception], typing.Union[Response, typing.Awaitable[Response]]
]
ExceptionHandlers = typing.Mapping[
    typing.Union[typing.Type[Exception], int], ExceptionHandler
]


class App:
    router: Router
    container: Container
    dependency_overrides: DependencyOverrideManager

    __slots__ = (
        "_container_state",
        "_debug",
        "_openapi_info",
        "_openapi_servers",
        "_openapi_version",
        "_openapi",
        "_root_path",
        "_root_path_in_servers",
        "_setup_run",
        "container",
        "dependency_overrides",
        "router",
    )

    def __init__(
        self,
        routes: typing.Optional[typing.Sequence[BaseRoute]] = None,
        *,
        container: typing.Optional[Container] = None,
        dependencies: typing.Optional[
            typing.Iterable[typing.Union[DependantBase[typing.Any], Depends]]
        ] = None,
        debug: bool = False,
        middleware: typing.Optional[typing.Sequence[Middleware]] = None,
        exception_handlers: typing.Optional[ExceptionHandlers] = None,
        lifespan: typing.Optional[
            typing.Callable[..., typing.AsyncContextManager[None]]
        ] = None,
        include_in_schema: bool = True,
        openapi_version: str = "3.0.3",
        title: str = "API",
        description: typing.Optional[str] = None,
        version: str = "0.1.0",
        openapi_url: typing.Optional[str] = "/openapi.json",
        docs_url: typing.Optional[str] = "/docs",
        servers: typing.Optional[typing.Iterable[openapi_models.Server]] = None,
        root_path: str = "",
        root_path_in_servers: bool = True,
    ) -> None:
        self.container = container or Container()
        _register_framework_dependencies(self.container, app=self)
        self.dependency_overrides = DependencyOverrideManager(self.container)
        self._container_state: ContainerState = ContainerState()
        self._setup_run = False

        @contextlib.asynccontextmanager
        async def lifespan_ctx(*_: typing.Any) -> typing.AsyncIterator[None]:
            lifespans, lifespan_deps = self._setup()
            self._setup_run = True
            original_container = self.container
            async with self._container_state.enter_scope(
                "app"
            ) as self._container_state:
                if lifespan is not None:
                    dep = Dependant(
                        _wrap_lifespan_as_async_generator(lifespan), scope="app"
                    )
                else:

                    async def null_lifespan() -> typing.AsyncIterator[None]:
                        yield

                    dep = Dependant(null_lifespan, scope="app")
                solved = self.container.solve(
                    JoinedDependant(
                        dep,
                        siblings=[
                            *(
                                Dependant(lifespan, scope="app")
                                for lifespan in lifespans
                            ),
                            *lifespan_deps,
                        ],
                    ),
                    scopes=Scopes,
                )
                try:
                    await self.container.execute_async(
                        solved, executor=AsyncExecutor(), state=self._container_state
                    )
                    yield
                finally:
                    # make this cm reentrant for testing purposes
                    self.container = original_container
                    self._setup_run = False
                    self._container_state = ContainerState()

        self._debug = debug

        routes = list(routes or [])
        routes.extend(
            self._get_doc_routes(
                openapi_url=openapi_url,
                docs_url=docs_url,
            )
        )
        middleware = _include_error_middleware(
            debug=debug,
            user_middleware=middleware or (),
            exception_handlers=exception_handlers or {},
        )
        self.router = Router(
            routes,
            dependencies=dependencies,
            middleware=middleware,
            include_in_schema=include_in_schema,
            lifespan=lifespan_ctx,
        )

        self._openapi_version = openapi_version
        self._openapi_info = openapi_models.Info(
            title=title,
            version=version,
            description=description,
        )
        self._openapi_servers = servers or []
        self._openapi: "typing.Optional[typing.Dict[str, typing.Any]]" = None
        self._root_path_in_servers = root_path_in_servers
        self._root_path = root_path

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        if self._root_path:
            prefix = scope.get("root_path", None)
            if prefix:
                scope["root_path"] = prefix.rstrip("/") + self._root_path
            else:
                scope["root_path"] = self._root_path
        scope_type = scope["type"]
        if scope_type == "lifespan":
            await self.router(scope, receive, send)
            return
        # http or websocket
        if not self._setup_run:
            self._setup()
        if "extensions" not in scope:
            scope["extensions"] = extensions = {}
        else:
            extensions = scope["extensions"]
        if scope_type == "http":
            if "xpresso" not in extensions:
                extensions["xpresso"] = XpressoHTTPExtension(
                    di_state=self._container_state
                )
        else:  # websocket
            if "xpresso" not in extensions:
                extensions["xpresso"] = XpressoWebSocketExtension(
                    di_state=self._container_state
                )
        await self.router(scope, receive, send)

    def _setup(
        self,
    ) -> typing.Tuple[
        typing.List[typing.Callable[..., typing.AsyncIterator[None]]],
        typing.List[DependantBase[typing.Any]],
    ]:
        lifespans: typing.List[typing.Callable[..., typing.AsyncIterator[None]]] = []
        lifespan_dependants: typing.List[DependantBase[typing.Any]] = []
        seen_routers: typing.Set[typing.Any] = set()
        for route in visit_routes(
            app_type=App, router=self.router, nodes=[self, self.router], path=""
        ):
            dependencies: typing.List[DependantBase[typing.Any]] = []
            for node in route.nodes:
                if isinstance(node, Router):
                    dependencies.extend(node.dependencies)
                    if node in seen_routers:
                        continue
                    seen_routers.add(node)
                    # avoid circular lifespan calls
                    if node is not self.router and node.lifespan is not None:
                        lifespans.append(
                            _wrap_lifespan_as_async_generator(node.lifespan)
                        )
            if isinstance(route.route, Path):
                for operation in route.route.operations.values():
                    operation.solve(
                        dependencies=[
                            *dependencies,
                            *route.route.dependencies,
                            *operation.dependencies,
                        ],
                        container=self.container,
                    )
                    for dep in operation.dependant.get_flat_subdependants():
                        if dep.scope == "app":
                            lifespan_dependants.append(dep)
            elif isinstance(route.route, WebSocketRoute):
                route.route.solve(
                    dependencies=[
                        *dependencies,
                        *route.route.dependencies,
                    ],
                    container=self.container,
                )
                for dep in route.route.dependant.get_flat_subdependants():
                    if dep.scope == "app":
                        lifespan_dependants.append(dep)
        return lifespans, lifespan_dependants

    def get_openapi(
        self, servers: typing.List[openapi_models.Server]
    ) -> openapi_models.OpenAPI:
        return genrate_openapi(
            visitor=visit_routes(
                app_type=App, router=self.router, nodes=[self, self.router], path=""
            ),
            container=self.container,
            version=self._openapi_version,
            info=self._openapi_info,
            servers=servers,
        )

    def _get_doc_routes(
        self,
        openapi_url: typing.Optional[str],
        docs_url: typing.Optional[str],
    ) -> typing.Iterable[BaseRoute]:
        routes: typing.List[BaseRoute] = []

        if openapi_url:
            openapi_url = openapi_url

            async def openapi(req: Request) -> JSONResponse:
                # get the root_path from the request and not just App._root_path
                # so that we can use the value set by the ASGI server
                # since ASGI servers also let you configure this
                root_path: str = req.scope.get("root_path", "").rstrip("/")  # type: ignore
                if self._openapi is None:
                    servers = list(self._openapi_servers)
                    if self._root_path_in_servers and root_path:
                        server_urls = {s.url for s in servers}
                        if root_path not in server_urls:
                            servers.insert(0, openapi_models.Server(url=root_path))
                    self._openapi = self.get_openapi(servers=servers).dict(
                        exclude_none=True, by_alias=True
                    )
                return JSONResponse(self._openapi)

            routes.append(
                StarletteRoute(
                    path=openapi_url, endpoint=openapi, include_in_schema=False
                )
            )
        if openapi_url and docs_url:

            openapi_url = openapi_url

            async def swagger_ui_html(req: Request) -> HTMLResponse:
                # see above for note on why we get root_path from the request
                root_path: str = req.scope.get("root_path", "").rstrip("/")  # type: ignore  # for Pylance
                full_openapi_url = root_path + openapi_url  # type: ignore[operator]
                return get_swagger_ui_html(
                    openapi_url=full_openapi_url,
                    title=f"{self._openapi_info.title} - Swagger UI",
                    oauth2_redirect_url=None,
                    init_oauth=None,
                )

            routes.append(
                StarletteRoute(
                    path=docs_url,
                    endpoint=swagger_ui_html,
                    include_in_schema=False,
                )
            )

        return routes


def _include_error_middleware(
    debug: bool,
    user_middleware: typing.Iterable[Middleware],
    exception_handlers: ExceptionHandlers,
) -> typing.Sequence[Middleware]:
    # user's exception handlers come last so that they can override
    # the default exception handlers
    exception_handlers = {
        RequestValidationError: validation_exception_handler,
        HTTPException: http_exception_handler,
        **exception_handlers,
    }

    error_handler = None
    for key, value in exception_handlers.items():
        if key in (500, Exception):
            error_handler = value
        else:
            exception_handlers[key] = value

    return (
        Middleware(ServerErrorMiddleware, handler=error_handler, debug=debug),
        *user_middleware,
        Middleware(ExceptionMiddleware, handlers=exception_handlers, debug=debug),
    )


def _wrap_lifespan_as_async_generator(
    lifespan: typing.Callable[..., typing.AsyncContextManager[None]]
) -> typing.Callable[..., typing.AsyncIterator[None]]:
    async def gen(
        *args: typing.Any, **kwargs: typing.Any
    ) -> typing.AsyncIterator[None]:
        async with lifespan(*args, **kwargs):
            yield

    sig = inspect.signature(gen)
    sig = sig.replace(parameters=list(inspect.signature(lifespan).parameters.values()))
    setattr(gen, "__signature__", sig)

    return gen


def _register_framework_dependencies(container: Container, app: App):
    container.bind(
        bind_by_type(
            Dependant(Request, scope="connection", wire=False),
            Request,
        )
    )
    container.bind(
        bind_by_type(
            Dependant(
                HTTPConnection,
                scope="connection",
                wire=False,
            ),
            HTTPConnection,
        )
    )
    container.bind(
        bind_by_type(
            Dependant(
                WebSocket,
                scope="connection",
                wire=False,
            ),
            WebSocket,
        )
    )
    container.bind(
        bind_by_type(
            Dependant(
                BackgroundTasks,
                scope="connection",
                wire=False,
            ),
            BackgroundTasks,
        )
    )
    container.bind(
        bind_by_type(
            Dependant(
                lambda: app.container,
                scope="app",
                wire=False,
            ),
            Container,
            covariant=True,
        )
    )
    container.bind(
        bind_by_type(
            Dependant(
                lambda: app,
                scope="app",
                wire=False,
            ),
            App,
            covariant=True,
        )
    )



import asyncio

from starlette.concurrency import run_in_threadpool
from starlette.exceptions import ExceptionMiddleware as StarletteExceptionMiddleware
from starlette.requests import Request
from starlette.types import Message, Receive, Scope, Send

from xpresso._utils.asgi import XpressoHTTPExtension
from xpresso.exceptions import HTTPException


class ExceptionMiddleware(StarletteExceptionMiddleware):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)  # type: ignore
            return

        response_started = False

        async def sender(message: Message) -> None:
            nonlocal response_started

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, sender)  # type: ignore
        except Exception as exc:
            handler = None

            if isinstance(exc, HTTPException):
                handler = self._status_handlers.get(exc.status_code)  # type: ignore

            if handler is None:
                handler = self._lookup_exception_handler(exc)  # type: ignore

            if handler is None:
                raise exc

            if response_started:
                msg = "Caught handled exception, but response already started."
                raise RuntimeError(msg) from exc

            request = Request(scope, receive=receive)
            if asyncio.iscoroutinefunction(handler):  # type: ignore
                response = await handler(request, exc)  # type: ignore
            else:
                response = await run_in_threadpool(handler, request, exc)  # type: ignore
            extension: XpressoHTTPExtension = scope["extensions"]["xpresso"]
            extension.response = response
            await response(scope, receive, sender)
            extension.response_sent = True



import inspect
import typing

from di.api.dependencies import CacheKey, DependantBase
from di.dependant import Dependant, Marker

from xpresso._utils.compat import Protocol
from xpresso.binders.api import (
    BodyExtractor,
    OpenAPIBody,
    OpenAPIParameter,
    ParameterExtractor,
)

T = typing.TypeVar("T", covariant=True)


class SupportsMarker(Protocol[T]):
    def register_parameter(self, param: inspect.Parameter) -> T:
        ...


class ParameterBinder(Dependant[typing.Any]):
    def __init__(
        self,
        in_: str,
        openapi: OpenAPIParameter,
        extractor: ParameterExtractor,
    ):
        self.in_ = in_
        self.openapi = openapi
        super().__init__(call=extractor.extract, scope="connection")

    @property
    def cache_key(self) -> CacheKey:
        # As per the spec, parameters are identified
        # by name and location
        return (self.__class__, self.openapi.name, self.in_)


class ParameterBinderMarker(Marker):
    def __init__(
        self,
        *,
        in_: str,
        extractor_marker: SupportsMarker[ParameterExtractor],
        openapi_marker: SupportsMarker[OpenAPIParameter],
    ) -> None:
        self.in_ = in_
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> DependantBase[typing.Any]:
        return ParameterBinder(
            in_=self.in_,
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )


class BodyBinder(Dependant[typing.Any]):
    def __init__(
        self,
        *,
        openapi: OpenAPIBody,
        extractor: BodyExtractor,
    ) -> None:
        super().__init__(call=extractor.extract_from_request, scope="connection")
        self.openapi = openapi
        self.extractor = extractor

    @property
    def cache_key(self) -> CacheKey:
        # Treat each body as unique since there should only be one body
        return (self.__class__, id(self))


class BodyBinderMarker(Marker):
    def __init__(
        self,
        *,
        extractor_marker: SupportsMarker[BodyExtractor],
        openapi_marker: SupportsMarker[OpenAPIBody],
    ) -> None:
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> BodyBinder:
        return BodyBinder(
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )



from __future__ import annotations

import typing

from di.api.providers import CallableProvider, CoroutineProvider
from di.dependant import Dependant

from xpresso.dependencies.models import Depends

T = typing.TypeVar("T")

Endpoint = typing.Union[CallableProvider[typing.Any], CoroutineProvider[typing.Any]]


class EndpointDependant(Dependant[typing.Any]):
    def __init__(
        self,
        endpoint: Endpoint,
        sync_to_thread: bool = False,
    ) -> None:
        super().__init__(
            call=endpoint,
            scope="endpoint",
            use_cache=False,
            wire=True,
            sync_to_thread=sync_to_thread,
        )

    def get_default_marker(self) -> Depends:
        return Depends()



import typing

from di.api.dependencies import DependantBase
from di.api.executor import SupportsAsyncExecutor
from di.api.solved import SolvedDependant
from di.container import Container
from di.dependant import JoinedDependant
from di.executors import AsyncExecutor, ConcurrentAsyncExecutor
from starlette.datastructures import URLPath
from starlette.requests import HTTPConnection, Request
from starlette.responses import JSONResponse, Response
from starlette.routing import BaseRoute, NoMatchFound, get_name
from starlette.types import ASGIApp, Receive, Scope, Send

import xpresso.binders.dependants as param_dependants
import xpresso.openapi.models as openapi_models
from xpresso._utils.asgi import XpressoHTTPExtension
from xpresso._utils.endpoint_dependant import Endpoint, EndpointDependant
from xpresso.dependencies.models import Depends, Scopes
from xpresso.encoders.api import Encoder
from xpresso.encoders.json import JsonableEncoder
from xpresso.responses import Responses


class _OperationApp:
    __slots__ = (
        "container",
        "dependant",
        "executor",
        "response_factory",
        "response_encoder",
    )

    def __init__(
        self,
        dependant: SolvedDependant[typing.Any],
        container: Container,
        executor: SupportsAsyncExecutor,
        response_factory: typing.Callable[[typing.Any], Response],
        response_encoder: Encoder,
    ) -> None:
        self.container = container
        self.dependant = dependant
        self.executor = executor
        self.response_factory = response_factory
        self.response_encoder = response_encoder

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        xpresso_scope: XpressoHTTPExtension = scope["extensions"]["xpresso"]
        if xpresso_scope.request is None:
            request = Request(scope=scope, receive=receive, send=send)
            xpresso_scope.request = request
        else:
            request = xpresso_scope.request
        values: typing.Dict[typing.Any, typing.Any] = {
            Request: request,
            HTTPConnection: request,
        }
        async with xpresso_scope.di_container_state.enter_scope(
            "connection"
        ) as cpnn_state:
            async with cpnn_state.enter_scope("endpoint") as endpoint_state:
                endpoint_return = await self.container.execute_async(
                    self.dependant,
                    values=values,
                    executor=self.executor,
                    state=endpoint_state,
                )
                if isinstance(endpoint_return, Response):
                    response = endpoint_return
                else:
                    response = self.response_factory(
                        self.response_encoder(endpoint_return)
                    )
                xpresso_scope.response = response
            await response(scope, receive, send)
            xpresso_scope.response_sent = True


class Operation(BaseRoute):
    def __init__(
        self,
        endpoint: Endpoint,
        *,
        # OpenAPI parameters
        include_in_schema: bool = True,
        tags: typing.Optional[typing.Sequence[str]] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        deprecated: typing.Optional[bool] = None,
        operation_id: typing.Optional[bool] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        external_docs: typing.Optional[openapi_models.ExternalDocumentation] = None,
        responses: typing.Optional[Responses] = None,
        # xpresso params
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[
            typing.Iterable[typing.Union[DependantBase[typing.Any], Depends]]
        ] = None,
        execute_dependencies_concurrently: bool = False,
        response_factory: typing.Callable[[typing.Any], Response] = JSONResponse,
        response_encoder: Encoder = JsonableEncoder(),
        sync_to_thread: bool = True,
    ) -> None:
        self._app: typing.Optional[ASGIApp] = None
        self.endpoint = endpoint
        self.tags = tuple(tags or ())
        self.summary = summary
        self.description = description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.servers = tuple(servers or ())
        self.external_docs = external_docs
        self.responses = dict(responses or {})
        self.dependencies = tuple(
            dep if not isinstance(dep, Depends) else dep.as_dependant()
            for dep in dependencies or ()
        )
        self.execute_dependencies_concurrently = execute_dependencies_concurrently
        self.response_factory = response_factory
        self.response_encoder = response_encoder
        self.include_in_schema = include_in_schema
        self.name: str = get_name(endpoint) if name is None else name  # type: ignore
        self.sync_to_thread = sync_to_thread

    async def handle(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        if self._app is None:
            raise RuntimeError("Operation cannot be used outside of a Xpresso App")
        return await self._app(scope, receive, send)

    def solve(
        self,
        container: Container,
        dependencies: typing.Iterable[typing.Union[DependantBase[typing.Any], Depends]],
    ) -> None:
        deps = [
            dep if not isinstance(dep, Depends) else dep.as_dependant()
            for dep in dependencies or ()
        ]
        self.dependant = container.solve(
            JoinedDependant(
                EndpointDependant(self.endpoint, sync_to_thread=self.sync_to_thread),
                siblings=[*deps, *self.dependencies],
            ),
            scopes=Scopes,
        )
        flat = self.dependant.get_flat_subdependants()
        bodies = [dep for dep in flat if isinstance(dep, param_dependants.BodyBinder)]
        if len(bodies) > 1:
            raise ValueError("There can only be 1 top level body per operation")
        executor: SupportsAsyncExecutor
        if self.execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self._app = _OperationApp(
            container=container,
            dependant=self.dependant,
            executor=executor,
            response_encoder=self.response_encoder,
            response_factory=self.response_factory,
        )

    def url_path_for(self, name: str, **path_params: str) -> URLPath:
        if path_params:
            raise NoMatchFound(name, path_params)
        if name != self.name:
            raise NoMatchFound(name, path_params)
        return URLPath("/")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, type(self)) and self.endpoint is __o.endpoint



import typing

import starlette.routing
import starlette.types
from di.api.dependencies import DependantBase
from di.api.providers import DependencyProvider as Endpoint

import xpresso.binders.dependants as param_dependants
import xpresso.openapi.models as openapi_models
from xpresso.dependencies.models import Depends
from xpresso.responses import Responses
from xpresso.routing.operation import Operation


class _PathApp:
    """Thin class wrapper so that Starlette treats us as an ASGI App"""

    __slots__ = ("operations",)

    def __init__(self, operations: typing.Mapping[str, Operation]) -> None:
        self.operations = operations

    def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> typing.Awaitable[None]:
        return self.operations[scope["method"]].handle(scope, receive, send)


class Path(starlette.routing.Route):
    include_in_schema: bool

    def __init__(
        self,
        path: str,
        *,
        get: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        head: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        post: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        put: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        patch: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        delete: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        connect: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        options: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        trace: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        redirect_slashes: bool = True,
        dependencies: typing.Optional[
            typing.Iterable[typing.Union[DependantBase[typing.Any], Depends]]
        ] = None,
        # OpenAPI metadata
        include_in_schema: bool = True,
        name: typing.Optional[str] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        parameters: typing.Optional[
            typing.Sequence[param_dependants.ParameterBinderMarker]
        ] = None,
        responses: typing.Optional[Responses] = None,
        tags: typing.Optional[typing.Iterable[str]] = None,
    ) -> None:
        if not path.startswith("/"):
            raise ValueError("Routed paths must start with '/'")
        self.path = path
        self.redirect_slashes = redirect_slashes
        self.dependencies = tuple(
            dep if not isinstance(dep, Depends) else dep.as_dependant()
            for dep in dependencies or ()
        )
        self.summary = summary
        self.description = description
        self.servers = tuple(servers or ())
        self.parameters = list(parameters or ())
        self.tags = tuple(tags or ())
        self.responses = dict(responses or {})

        operations: typing.Dict[str, Operation] = {}
        if get:
            operations["GET"] = get if isinstance(get, Operation) else Operation(get)
        if head:
            operations["HEAD"] = (
                head if isinstance(head, Operation) else Operation(head)
            )
        if post:
            operations["POST"] = (
                post if isinstance(post, Operation) else Operation(post)
            )
        if put:
            operations["PUT"] = put if isinstance(put, Operation) else Operation(put)
        if patch:
            operations["PATCH"] = (
                patch if isinstance(patch, Operation) else Operation(patch)
            )
        if delete:
            operations["DELETE"] = (
                delete if isinstance(delete, Operation) else Operation(delete)
            )
        if connect:
            operations["CONNECT"] = (
                connect if isinstance(connect, Operation) else Operation(connect)
            )
        if options:
            operations["OPTIONS"] = (
                options if isinstance(options, Operation) else Operation(options)
            )
        if trace:
            operations["TRACE"] = (
                trace if isinstance(trace, Operation) else Operation(trace)
            )
        self.operations = operations
        super().__init__(  # type: ignore  # for Pylance
            path=path,
            endpoint=_PathApp(operations),
            name=name,  # type: ignore[arg-type]
            include_in_schema=include_in_schema,
            methods=list(operations.keys()),
        )



import typing
from dataclasses import dataclass

from starlette.routing import BaseRoute, Mount
from starlette.routing import Router as StarletteRouter

from xpresso._utils.compat import Protocol
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router as XpressoRouter
from xpresso.routing.websockets import WebSocketRoute

Router = typing.Union[XpressoRouter, StarletteRouter]


class App(Protocol):
    @property
    def router(self) -> XpressoRouter:
        ...


AppType = typing.TypeVar("AppType", bound=App)


@dataclass(frozen=True)
class VisitedRoute(typing.Generic[AppType]):
    path: str
    nodes: typing.List[typing.Union[Router, AppType]]
    route: BaseRoute


def visit_routes(
    app_type: typing.Type[AppType],
    router: Router,
    nodes: typing.List[typing.Union[Router, AppType]],
    path: str,
) -> typing.Generator[VisitedRoute[AppType], None, None]:
    for route in typing.cast(typing.Iterable[BaseRoute], router.routes):  # type: ignore  # for Pylance
        if isinstance(route, Mount):
            app: typing.Any = route.app
            mount_path: str = route.path  # type: ignore  # for Pylance
            if isinstance(app, (StarletteRouter, XpressoRouter)):
                yield VisitedRoute(
                    path=path,
                    nodes=nodes + [app],
                    route=route,
                )
                yield from visit_routes(
                    app_type=app_type,
                    router=app,
                    nodes=nodes + [app],
                    path=path + mount_path,
                )
            elif isinstance(app, app_type):
                yield VisitedRoute(
                    path=path,
                    nodes=nodes + [app, app.router],
                    route=route,
                )
                yield from visit_routes(
                    app_type=app_type,
                    router=app.router,
                    nodes=nodes + [app, app.router],
                    path=path + mount_path,
                )
        elif isinstance(route, (Path, WebSocketRoute)):
            yield VisitedRoute(
                path=path + route.path,
                nodes=nodes,
                route=route,
            )



import inspect
import typing

from di.typing import get_markers_from_annotation

from xpresso._utils.compat import Annotated, get_args, get_origin
from xpresso._utils.typing import model_field_from_param
from xpresso.binders.api import ModelNameMap, OpenAPIBody, Schemas
from xpresso.binders.dependants import BodyBinderMarker
from xpresso.openapi import models as openapi_models


class OpenAPIContentTypeDiscriminated(typing.NamedTuple):
    sub_body_providers: typing.Iterable[OpenAPIBody]
    description: typing.Optional[str]
    required: typing.Optional[bool]
    include_in_schema: bool

    def get_models(self) -> typing.List[type]:
        return [model for b in self.sub_body_providers for model in b.get_models()]

    def get_openapi_body(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.RequestBody:
        content: typing.Dict[str, openapi_models.MediaType] = {}
        for body in self.sub_body_providers:
            content.update(body.get_openapi_body(model_name_map, schemas).content)
        return openapi_models.RequestBody(
            description=self.description,
            required=self.required,
            content=content,
        )

    def get_field_encoding(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Encoding:
        raise NotImplementedError

    def get_field_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Schema:
        raise NotImplementedError


class OpenAPIContentTypeDiscriminatedMarker(typing.NamedTuple):
    description: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        field = model_field_from_param(param)
        required = field.required is not False
        sub_body_providers: typing.List[OpenAPIBody] = []

        annotation = param.annotation
        origin = get_origin(annotation)
        assert origin is Annotated
        annotation = next(iter(get_args(annotation)))
        origin = get_origin(annotation)
        if origin is not typing.Union:
            raise TypeError("Unioned bodies must be a Union of simple bodies")
        args = get_args(annotation)
        for arg in args:
            sub_body_param = inspect.Parameter(
                name=param.name,
                kind=param.kind,
                annotation=arg,
                default=param.default,
            )
            marker = next(get_markers_from_annotation(arg, BodyBinderMarker), None)
            if marker is None:
                raise TypeError(f"Type annotation is missing body marker: {arg}")
            sub_body_openapi = marker.openapi_marker
            provider = sub_body_openapi.register_parameter(sub_body_param)
            if provider.include_in_schema:
                sub_body_providers.append(provider)
        return OpenAPIContentTypeDiscriminated(
            sub_body_providers=sub_body_providers,
            description=self.description,
            required=None if required else False,
            include_in_schema=True,
        )



from typing import Any, Dict, Iterable, List, Optional, Union

from starlette.datastructures import UploadFile
from starlette.requests import HTTPConnection, Request

import xpresso.openapi.models as openapi_models
from xpresso._utils.compat import Protocol


class ParameterExtractor(Protocol):
    def extract(self, connection: HTTPConnection) -> Any:
        raise NotImplementedError


class BodyExtractor(Protocol):
    async def extract_from_request(self, request: Request) -> Any:
        """Extract from top level request"""
        raise NotImplementedError

    def matches_media_type(self, media_type: Optional[str]) -> bool:
        """Check if this extractor can extract the given media type"""
        raise NotImplementedError

    # APIs to support being a field in a FormData or Multipart request
    async def extract_from_field(
        self,
        field: Union[str, UploadFile],
        *,
        loc: Iterable[Union[str, int]],
    ) -> Any:
        """Extract from a form field"""
        raise NotImplementedError


Model = type
ModelNameMap = Dict[Model, str]
Schemas = Dict[str, Union[openapi_models.Schema, openapi_models.Reference]]


class _SupportsGetModels(Protocol):
    @property
    def include_in_schema(self) -> bool:
        ...

    def get_models(self) -> List[type]:
        """Collect all of the types that OpenAPI schemas will be
        produced from.

        Xpresso will then assign a schema name to each type and pass
        that back via the ModelNameMap parameter.

        This ensures that all schema models have a unique name,
        even if their Python class names conflict.
        """
        raise NotImplementedError


class OpenAPIParameter(_SupportsGetModels, Protocol):
    @property
    def name(self) -> str:
        ...

    @property
    def in_(self) -> openapi_models.ParameterLocations:
        ...

    def get_openapi_parameter(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.ConcreteParameter:
        """Generate the OpenAPI spec for this parameter"""
        raise NotImplementedError


class OpenAPIBody(_SupportsGetModels, Protocol):
    def get_openapi_body(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.RequestBody:
        """Generate the OpenAPI spec for this request body"""
        raise NotImplementedError

    # APIs to support being a field in a FormData or Multipart request
    def get_field_encoding(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Encoding:
        """Get the Encoding for this field.

        If this is a URL-encoded form field (as part of
        an application/x-www-form-urlencoded) the Encoding
        MAY include the `style` and `explode` sections.

        For all other field types, it can only include the
        `contentType` and `Headers` section.
        """
        raise NotImplementedError

    def get_field_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Schema:
        """Get the Schema for this field.

        This will be used to assemble the Form's schema,
        which is an object schema having as properties the schemas
        for each field.
        """
        raise NotImplementedError



import inspect
import typing

from starlette.datastructures import FormData

import xpresso.openapi.models as openapi_models
from xpresso._utils.compat import Protocol
from xpresso._utils.typing import Some
from xpresso.binders.api import ModelNameMap, Schemas


class FormFieldOpenAPIProvider(Protocol):
    @property
    def field_name(self) -> str:
        ...

    @property
    def include_in_schema(self) -> bool:
        ...

    def get_models(self) -> typing.List[type]:
        raise NotImplementedError

    def get_field_encoding(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Encoding:
        raise NotImplementedError

    def get_field_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Schema:
        raise NotImplementedError


class FormDataFieldOpenapiMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> FormFieldOpenAPIProvider:
        raise NotImplementedError


class FormDataExtractor(Protocol):
    @property
    def field_name(self) -> str:
        ...

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[int, str]]
    ) -> typing.Optional[Some[typing.Any]]:
        raise NotImplementedError


class FormDataExtractorMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> FormDataExtractor:
        raise NotImplementedError


class FormFieldMarker(typing.NamedTuple):
    openapi_marker: FormDataFieldOpenapiMarker
    extractor_marker: FormDataExtractorMarker



"""Form data deserialization as per https://swagger.io/docs/specification/serialization/#query
"""
import functools
import re
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from pydantic.fields import ModelField
from starlette.datastructures import UploadFile

from xpresso._utils.compat import Protocol
from xpresso._utils.typing import Some, is_mapping_like, is_sequence_like
from xpresso.binders._utils.grouped import grouped
from xpresso.binders.exceptions import InvalidSerialization


class UnexpectedFileReceived(TypeError):
    pass


def get_matches(
    params: Iterable[Tuple[str, Union[str, UploadFile]]], name: str
) -> List[Optional[str]]:
    matches: List[Optional[str]] = []
    for k, v in params:
        if k == name:
            if not isinstance(v, str):
                raise UnexpectedFileReceived(
                    "Expected a string form field but received a file"
                )
            matches.append(v or None)  # convert "" (from param=&other=123) to None
    return matches


def collect_form_sequence(
    params: Iterable[Tuple[str, Union[str, UploadFile]]],
    name: str,
    explode: bool,
    delimiter: str,
) -> Optional[Some[List[Optional[str]]]]:
    matches = get_matches(params, name)
    if not matches:
        return Some([])
    if explode:
        # No further processing needed since the delimiter is &
        # for all styles
        return Some(matches)
    else:
        # We need to split these up manually
        if len(matches) != 1:
            raise InvalidSerialization("Invalid form serialziation")
        match = matches[0]
        if not match:
            # user gave us ?param=&other=abc
            return Some([None])
        return Some(list(match.split(delimiter)))


def collect_object(
    params: Iterable[Tuple[str, Union[str, UploadFile]]],
    name: str,
    explode: bool,
) -> Optional[Some[Dict[str, str]]]:
    if explode:
        # free form params, let validation filter them out
        return Some({k: v for k, v in params if isinstance(v, str)})
    else:
        matches = get_matches(params, name)
        if not matches:
            return None
        if len(matches) != 1:
            raise InvalidSerialization("Invalid form serialziation")
        match = matches[0]
        if not match:
            return Some({})
        return Some(dict(grouped(match.split(","))))  # type: ignore[arg-type]


def collect_deep_object(
    params: Iterable[Tuple[str, Union[str, UploadFile]]], name: str
) -> Optional[Some[Optional[Dict[str, Any]]]]:
    # deepObject does not support repeated fields so we can put our fields in dict
    param_dict: Dict[str, str] = {k: v for k, v in params if isinstance(v, str)}
    if not param_dict:
        return None
    res: Dict[str, Any] = {}
    path_regex = re.compile(rf"{name}\[(\w+)\]")
    for path, value in param_dict.items():
        match = path_regex.match(path)
        if not match:
            continue
        field_name = match.group(1)
        res[field_name] = value
    return Some(res or None)


def collect_scalar(
    params: Iterable[Tuple[str, Union[str, UploadFile]]], name: str
) -> Optional[Some[Optional[str]]]:
    params_mapping = dict(params)
    if name not in params_mapping:
        return None
    v = params_mapping[name]
    if not isinstance(v, str):
        raise UnexpectedFileReceived("Expected a string form field but received a file")
    return Some(v or None)


delimiters = {
    "form": ",",
    "pipeDelimited": "|",
    "spaceDelimited": " ",
}


class Extractor(Protocol):
    def __call__(
        self, *, name: str, params: Iterable[Tuple[str, Union[str, UploadFile]]]
    ) -> Optional[Some[Any]]:
        ...


def get_extractor(*, style: str, explode: bool, field: ModelField) -> Extractor:
    if style == "deepObject":
        return collect_deep_object
    if style in ("spaceDelimited", "pipeDelimited"):
        return functools.partial(
            collect_form_sequence,
            explode=explode,
            delimiter=delimiters[style],
        )
    # form style
    if is_sequence_like(field):
        return functools.partial(
            collect_form_sequence,
            explode=explode,
            delimiter=delimiters[style],
        )
    if is_mapping_like(field):
        return functools.partial(
            collect_object,
            explode=explode,
        )
    return collect_scalar



from typing import Any, Iterable, Mapping, Optional, Union

from pydantic import BaseModel, Field
from starlette.responses import (
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    Response,
    StreamingResponse,
)

from xpresso._utils.asgi import XpressoHTTPExtension
from xpresso.exceptions import XpressoError
from xpresso.openapi.models import Example, ResponseHeader
from xpresso.requests import Request

__all__ = (
    "get_response",
    "set_response",
    "ResponseSpec",
    "JsonResponseSpec",
    "FileResponseSpec",
    "HTMLResponseSpec",
    "PlainTextResponseSpec",
    "Response",
    "JSONResponse",
    "StreamingResponse",
    "HTMLResponse",
    "PlainTextResponse",
)


def get_response(request: Request) -> Response:
    xpresso_extension: XpressoHTTPExtension = request.scope["extensions"]["xpresso"]  # type: ignore  # for Pylance
    if xpresso_extension.response is None:
        raise LookupError(
            "xpresso.responses.get_response was called"
            " before the endpoint has finished executing or the endpoint raised an exception"
        )
    return xpresso_extension.response


def set_response(request: Request, response: Response) -> None:
    xpresso_extension: XpressoHTTPExtension = request.scope["extensions"]["xpresso"]  # type: ignore  # for Pylance
    if xpresso_extension.response_sent:
        raise XpressoError(
            'set_response() can only be used from "endpoint" scoped dependendencies'
        )
    xpresso_extension.response = response


class ResponseSpec(BaseModel):
    description: str
    model: Any = None
    media_type: Optional[str] = None
    headers: Optional[Mapping[str, Union[ResponseHeader, str]]] = None
    examples: Optional[Mapping[str, Union[Example, Any]]] = None


class JsonResponseSpec(ResponseSpec):
    model: Any = Field(default={})
    media_type: str = "application/json"


class FileResponseSpec(ResponseSpec):
    model: Any = Field(default=bytes)
    media_type: str = "application/octetstream"


class HTMLResponseSpec(ResponseSpec):
    model: Any = Field(default=str)
    media_type: str = "text/html"


class PlainTextResponseSpec(ResponseSpec):
    model: str
    media_type = "text/plain"


Responses = Mapping[Union[int, str], Union[Iterable[ResponseSpec], ResponseSpec]]



import json
import typing

import xpresso.binders.dependants as param_dependants
import xpresso.openapi.models as openapi_models
from xpresso._utils.compat import Annotated, Literal
from xpresso.binders._body.extractors.discriminated import (
    ContentTypeDiscriminatedExtractorMarker,
)
from xpresso.binders._body.extractors.file import FileBodyExtractorMarker
from xpresso.binders._body.extractors.form import FormDataBodyExtractorMarker
from xpresso.binders._body.extractors.form_encoded_field import (
    FormEncodedFieldExtractorMarker,
)
from xpresso.binders._body.extractors.form_field_extractor import FieldExtractorMarker
from xpresso.binders._body.extractors.json import Decoder, JsonBodyExtractorMarker
from xpresso.binders._body.form_field import FormFieldMarker
from xpresso.binders._body.openapi.discriminated import (
    OpenAPIContentTypeDiscriminatedMarker,
)
from xpresso.binders._body.openapi.file import OpenAPIFileMarker
from xpresso.binders._body.openapi.form import OpenAPIFormDataMarker
from xpresso.binders._body.openapi.form_encoded_field import OpenAPIFormFieldMarker
from xpresso.binders._body.openapi.form_field_extractor import OpenAPIFieldMarker
from xpresso.binders._body.openapi.json import OpenAPIJsonMarker

Example = typing.Union[openapi_models.Example, typing.Any]


def Json(
    *,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    decoder: Decoder = json.loads,
    enforce_media_type: bool = True,
    consume: bool = True,
    include_in_schema: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = JsonBodyExtractorMarker(
        decoder=decoder,
        enforce_media_type=enforce_media_type,
        consume=consume,
    )
    openapi = OpenAPIJsonMarker(
        description=description,
        examples=examples,
        include_in_schema=include_in_schema,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def File(
    *,
    media_type: typing.Optional[str] = None,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    format: Literal["binary", "base64"] = "binary",
    consume: bool = True,
    include_in_schema: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = FileBodyExtractorMarker(
        media_type=media_type,
        enforce_media_type=enforce_media_type,
        consume=consume,
    )
    openapi = OpenAPIFileMarker(
        description=description,
        examples=examples,
        media_type=media_type,
        format=format,
        include_in_schema=include_in_schema,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def FormEncodedField(
    *,
    alias: typing.Optional[str] = None,
    style: openapi_models.FormDataStyles = "form",
    explode: bool = True,
    include_in_schema: bool = True,
) -> FormFieldMarker:
    extractor = FormEncodedFieldExtractorMarker(
        alias=alias,
        style=style,
        explode=explode,
    )
    openapi = OpenAPIFormFieldMarker(
        alias=alias,
        style=style,
        explode=explode,
        include_in_schema=include_in_schema,
    )
    return FormFieldMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def FormField(
    *,
    alias: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> FormFieldMarker:
    extractor = FieldExtractorMarker(alias=alias, repeated=False)
    openapi = OpenAPIFieldMarker(
        alias=alias,
        include_in_schema=include_in_schema,
        repeated=False,
    )
    return FormFieldMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def RepeatedFormField(
    *,
    alias: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> FormFieldMarker:
    extractor = FieldExtractorMarker(alias=alias, repeated=True)
    openapi = OpenAPIFieldMarker(
        alias=alias,
        include_in_schema=include_in_schema,
        repeated=True,
    )
    return FormFieldMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def Form(
    *,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = FormDataBodyExtractorMarker(
        media_type="application/x-www-form-urlencoded",
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIFormDataMarker(
        description=description,
        examples=examples,
        media_type="application/x-www-form-urlencoded",
        include_in_schema=include_in_schema,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def Multipart(
    *,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = FormDataBodyExtractorMarker(
        media_type="multipart/form-data",
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIFormDataMarker(
        description=description,
        examples=examples,
        media_type="multipart/form-data",
        include_in_schema=include_in_schema,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def ContentTypeDiscriminatedBody(
    *,
    description: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = ContentTypeDiscriminatedExtractorMarker()
    openapi = OpenAPIContentTypeDiscriminatedMarker(
        description=description,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


# Convenience type aliases
_T = typing.TypeVar("_T")
FromJson = Annotated[_T, Json()]
FromFile = Annotated[_T, File()]
FromFormField = Annotated[_T, FormEncodedField()]
ExtractField = Annotated[_T, FormField()]
ExtractRepeatedField = Annotated[_T, RepeatedFormField()]
FromFormData = Annotated[_T, Form()]
FromMultipart = Annotated[_T, Multipart()]
ByContentType = Annotated[_T, ContentTypeDiscriminatedBody()]



import sys

if sys.version_info < (3, 9):
    from typing_extensions import Annotated as Annotated  # noqa: F401
    from typing_extensions import get_args as get_args  # noqa: F401
    from typing_extensions import get_origin as get_origin  # noqa: F401
    from typing_extensions import get_type_hints as get_type_hints  # noqa: F401
else:
    from typing import Annotated as Annotated  # noqa: F401
    from typing import get_args as get_args  # noqa: F401
    from typing import get_origin as get_origin  # noqa: F401
    from typing import get_type_hints as get_type_hints  # noqa: F401

if sys.version_info < (3, 8):
    from typing_extensions import Literal as Literal  # noqa: F401
    from typing_extensions import Protocol as Protocol  # noqa: F401
    from typing_extensions import TypedDict as TypedDict  # noqa: F401
    from typing_extensions import runtime_checkable as runtime_checkable  # noqa: F401
else:

    from typing import Literal as Literal  # noqa: F401
    from typing import Protocol as Protocol  # noqa: F401
    from typing import runtime_checkable as runtime_checkable  # noqa: F401

    from typing_extensions import TypedDict as TypedDict  # noqa: F401
